'''
Created on 31-05-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
from math import ceil
import random
from aqopa.simulator.state import Hook, ExecutionResult
from aqopa.model import AssignmentInstruction,\
    CallFunctionInstruction, IfInstruction, WhileInstruction,\
    CommunicationInstruction, CallFunctionExpression, TupleExpression, ComparisonExpression
from aqopa.module.timeanalysis.error import TimeSynchronizationException
from aqopa.simulator.error import RuntimeException

class PreInstructionHook(Hook):
    """
    Execution hook executed before default core execution of each instruction.
    Returns execution result.
    """
    
    def __init__(self, module, simulator):
        """ """
        self.module = module
        self.simulator = simulator
        
        
    def execute(self, context):
        """
        """
        instruction = context.get_current_instruction()
        
        if instruction.__class__ not in [AssignmentInstruction, CallFunctionInstruction, 
                                     IfInstruction, WhileInstruction, CommunicationInstruction]:
            return
        
        if isinstance(instruction, CommunicationInstruction):
            return self._execute_communication_instruction(context)
        else:
            self._update_time(context)
            return ExecutionResult()
            
    def _update_time(self, context):
        """ 
        Update times in context according to current instruction.
        """
        instruction = context.get_current_instruction()
        
        expression = None
        if isinstance(instruction, AssignmentInstruction):
            expression = instruction.expression
        elif isinstance(instruction, CallFunctionInstruction):
            expression = CallFunctionExpression(instruction.function_name, 
                                                instruction.arguments, 
                                                instruction.qop_arguments)
        else:
            expression = instruction.condition
            
        # Return details for each expression in instruction
        # In some instruction may be more expressions (tuple, nested call function)
        total_time, time_details = self._get_time_details_for_expression(context, expression)
        
        if total_time > 0:
            host = context.get_current_host()
            self.module.add_timetrace(self.simulator, host, host.get_current_process(),
                                      instruction, time_details,
                                      self.module.get_current_time(self.simulator, host),
                                      total_time)
            self.module.set_current_time(self.simulator, host, self.module.get_current_time(self.simulator, host) + total_time)
            
    def _get_time_details_for_expression(self, context, expression):
        """
        Returns calculated time that execution of expression takes.
        """
        if isinstance(expression, TupleExpression):
            return self._get_time_details_for_tuple_expression(context, expression)
        elif isinstance(expression, CallFunctionExpression):
            return self._get_time_details_for_simple_expression(context, expression)
        elif isinstance(expression, ComparisonExpression):
            return self._get_time_details_for_comparison_expression(context, expression)
        return 0, []

    def _get_time_details_for_tuple_expression(self, context, expression):
        """
        Calculate execution time for tuple expression. 
        """
        total_time = 0
        time_details = [] 
        for i in range(0, len(expression.elements)):
            element_total_time, element_time_details = self._get_time_details_for_expression(context, expression.elements[i])
            total_time += element_total_time
            time_details.extend(element_time_details)
        return total_time, time_details

    def _get_time_details_for_comparison_expression(self, context, expression):
        """
        Calculate execution time for tuple expression.
        """
        total_time = 0
        time_details = []

        element_total_time, element_time_details = self._get_time_details_for_expression(context, expression.left)
        total_time += element_total_time
        time_details.extend(element_time_details)

        element_total_time, element_time_details = self._get_time_details_for_expression(context, expression.right)
        total_time += element_total_time
        time_details.extend(element_time_details)

        return total_time, time_details
    
    def _get_time_details_for_simple_expression(self, context, expression):
        """
        Calculate time for expression.
        """
        time = 0
        time_details = []
        
        metric = context.metrics_manager\
            .find_primitive(context.get_current_host(), expression)

        if metric:
            block = metric.block
            
            for i in range(0, len(block.service_params)):
                sparam = block.service_params[i]
                
                if sparam.service_name.lower() != "time":
                    continue

                metric_type = sparam.param_name.lower()
                metric_unit = sparam.unit
                metric_value = metric.service_arguments[i]

                if metric_type == "exact":

                    if metric_unit == "ms":
                        time = float(metric_value)

                    elif metric_unit == "s":
                        time = float(metric_value) * 1000

                    elif metric_unit == "mspb" or metric_unit == "mspB":

                        mparts = metric_value.split(':')
                        if len(mparts) != 2:
                            raise RuntimeException(
                                'Metric unit is set as %s, but call argument to get size of is not set.'
                                % metric_unit)

                        size = 0
                        call_params_indexes = mparts[1].split(',')
                        for index in call_params_indexes:
                            index = int(index)
                            populated_expression = context.expression_populator.populate(
                                                        expression.arguments[index-1],
                                                        context.get_current_host().get_variables(),
                                                        context.expression_reducer)

                            size += context.metrics_manager.get_expression_size(
                                                                populated_expression,
                                                                context,
                                                                context.get_current_host())

                        msperbyte = float(mparts[0])
                        if metric_unit == "mspb":
                            msperbyte = msperbyte / 8.0

                        time = msperbyte * size

                elif metric_type == "range":

                    mvalues = metric_value.split('..')
                    val_from = float(mvalues[0])
                    val_to = float(mvalues[1])

                    time = val_from + (val_to-val_from)*random.random()

                elif metric_type == "block":

                    mparts = metric_value.split(':')
                    element_index = int(mparts[0])-1
                    unit_time = float(mparts[1])
                    unit_size = int(mparts[2])

                    #time_unit = metric_unit[0]
                    size_unit = metric_unit[1]

                    argument_size = context.metrics_manager.get_expression_size(expression.arguments[element_index],
                                                                                context,
                                                                                context.get_current_host())
                    units = ceil(argument_size / float(unit_size))

                    time = units * unit_time
                    if size_unit == 'b':
                        time *= 8.0

            if time > 0:
                time_details.append((expression, time))

        for expr in expression.arguments:
            argument_time, argument_time_details = self._get_time_details_for_expression(context, expr)
            
            time += argument_time
            time_details.extend(argument_time_details)

        return time, time_details
                
    def _execute_communication_instruction(self, context):
        """ """
        channel = context.channels_manager.find_channel_for_current_instruction(context)
        if not channel:
            return None
        
        instruction = context.get_current_instruction()
        expressions_cnt = len(instruction.variables_names)

        # If host does not want to receive or send message,
        # let it go
        if isinstance(instruction, CommunicationInstruction):

            # Check if other hosts should send or receive their message through this channel before
            current_host_time = self.module.get_current_time(self.simulator, context.get_current_host())
            min_hosts_time = (context.get_current_host(), current_host_time)
            delay_communication_execution = False

            for h in context.hosts:
                # Omit finished hosts
                if h.finished():
                    continue
                # Omit current host
                if h == context.get_current_host():
                    continue
                # Omit hosts not using current channel
                if not channel.connected_with_host(h):
                    continue

                # Check the time of other hosts
                host_time = self.module.get_current_time(self.simulator, h)
                # If checked (in loop) host is in the past relative to the time of actual host
                if host_time < min_hosts_time[1]:
                    current_instruction = h.get_current_instructions_context().get_current_instruction()
                    if isinstance(current_instruction, CommunicationInstruction):
                        # If the checked host (the one in the past) want to send message
                        # it should be first
                        if current_instruction.is_out():
                            delay_communication_execution = True
                        else:
                            # The checked host (the one in the past) is waiting for the message
                            # Lets check if the message fot the host is in the channel
                            # If there is, it should be executed first so wait for it
                            vars_cnt = len(current_instruction.variables_names)
                            channel = context.channels_manager.find_channel_for_host_instruction(
                                context, h, current_instruction)
                            # If channel has enough vars
                            if len(channel.get_queue_of_sending_hosts(vars_cnt)) >= vars_cnt:
                                delay_communication_execution = True
                    else:
                        # If the checked host (the one in the past) executes some non-communication operation
                        # it should be done first
                        delay_communication_execution = True


                    #ch = context.channels_manager.find_channel_for_host_instruction(
                    #    context, h, current_instruction)
                    #if ch == channel:
                    #    if current_instruction.is_out():
                    #        host_time = self.module.get_current_time(self.simulator, h)
                    #        if host_time < min_hosts_time[1]:
                    #            min_hosts_time = (h, host_time)
                    #
                    #ch = None
                    #if instruction.is_out(): # OUT instruction
                    #    if not current_instruction.is_out(): # IN instruction
                    #        ch = context.channels_manager.find_channel_for_host_instruction(
                    #            context, h, current_instruction)
                    #else: # IN instruction
                    #    if current_instruction.is_out(): # OUT instruction
                    #        ch = context.channels_manager.find_channel_for_host_instruction(
                    #            context, h, current_instruction)
                    #if ch == channel:
                    #    continue
                    #
                    #if self.module.get_current_time(self.simulator, h) < current_host_time:
                    #    delay_communication_execution = True

            ## Delay execution of this instruction
            ## if needed according to previous check
            #if delay_communication_execution:
            #    if context.get_current_host() != min_hosts_time[0]:
            #        return ExecutionResult(custom_index_management=True,
            #                               finish_instruction_execution=True)

            if delay_communication_execution:
                return ExecutionResult(custom_index_management=True,
                                       finish_instruction_execution=True)

            if instruction.is_out():
                sender = context.get_current_host()
                receivers_list = channel.get_queue_of_receiving_hosts(expressions_cnt)
                time_of_sending = self.module.get_current_time(self.simulator, sender)
                for i in range(0, len(receivers_list)):
                    if not receivers_list[i]:  # No receiver for message
                        self.module.add_channel_message_trace(self.simulator, channel,
                                                              self.module.get_channel_next_message_id(self.simulator,
                                                                                                      channel),
                                                              sender, time_of_sending)
                    else:
                        if self.module.get_current_time(self.simulator, sender) < self.module.get_current_time(self.simulator, receivers_list[i]):

                            raise TimeSynchronizationException("Time synchronization error. Trying to send message from host '%s' at time %s ms while receiving host '%s' has time %s ms." %
                                                               (unicode(sender),
                                                                self.module.get_current_time(self.simulator, sender),
                                                                unicode(receivers_list[i]),
                                                                self.module.get_current_time(self.simulator, receivers_list[i])))

                        # When someone is already waiting for the message
                        # the sent at time and received at time are the same
                        time_of_receiving = time_of_sending
                        self.module.add_channel_message_trace(self.simulator, channel,
                                                              self.module.get_channel_next_message_id(self.simulator, channel),
                                                              sender, time_of_sending,
                                                              receivers_list[i], time_of_receiving)
                        self.module.set_current_time(self.simulator, receivers_list[i], time_of_receiving)


            else: # IN communication step
                sender = None
                receiver = context.get_current_host()
                sending_hosts = channel.get_queue_of_sending_hosts(expressions_cnt)
                for sending_host in sending_hosts:
                    traces = self.module.get_channel_message_traces(self.simulator, channel)
                    for trace in traces:
                        if trace.sender == sending_host and not trace.receiver:
                            time_of_sending = self.module.get_current_time(self.simulator, sending_host)
                            time_of_receiving = self.module.get_current_time(self.simulator, receiver)
                            if time_of_sending < time_of_receiving:
                                raise TimeSynchronizationException(
                                    ("Time synchronization error. " +
                                     "Trying to send message from host '%s' at time %s ms " +
                                     "while receiving host '%s' has time %s ms.") %
                                    (unicode(sender), self.module.get_current_time(self.simulator, sender),
                                     unicode(trace.receiver),
                                     self.module.get_current_time(self.simulator, trace.receiver)))
                            # Calculate time of receiving as the maximum of sending and receiving
                            time_of_receiving = max(time_of_sending, time_of_receiving)
                            trace.add_receiver(receiver, time_of_receiving)
                            self.module.set_current_time(self.simulator, receiver, time_of_receiving)

                            
            
        


'''
Created on 07-05-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
from aqopa.model import original_name, name_indexes,\
    CommunicationInstruction
from aqopa.simulator.error import RuntimeException

class ChannelMessage():
    """
    Message sent through channel.
    """
    
    def __init__(self, sender, expression):
        self.sender = sender
        self.expression = expression
        
        
class ChannelMessageRequest():
    """
    Representation of hosts' requests for messages.
    """
    
    def __init__(self, receiver, communication_instruction):
        """ """
        self.receiver       = receiver
        self.instruction    = communication_instruction

class Channel():
    """
    Simulation channel.
    """
    
    def __init__(self, name, buffer_size):
        self.name = name
        self._buffer_size           = buffer_size   # Size of buffer, negative means that buffer is unlimited,
                                                    # zero means that channel is cynhronous 
        
        self._connected_hosts       = []
        self._connected_processes   = []
        
        self._waiting_requests      = []
        self._sent_messages         = []
        self._dropped_messages_cnt  = 0
        
    def _bind_sent_expressions_with_receivers(self):
        """
        Checks if have any messages and requests and if yes, bind them.
        """
        while len(self._waiting_requests) > 0:
            request = self._waiting_requests[0]
            if len(request.instruction.variables_names) > len(self._sent_messages):
                break
            
            size = len(request.instruction.variables_names)
            for i in range(0, size):
                message = self._sent_messages.pop(0)
                
                request.receiver.set_variable(request.instruction.variables_names[i],
                                              message.expression)
                
            request.receiver.get_instructions_context_of_instruction(request.instruction)\
                .goto_next_instruction()
            request.receiver.mark_changed()
            
            self._waiting_requests.pop(0)
        
    def clone(self):
        return Channel(self.original_name(), self._buffer_size)
        
    def original_name(self):
        """
        Return original name of channel - without indexes.
        """
        return original_name(self.name)
    
    def indexes(self):
        """
        Return name indexes of channel 
        """
        return name_indexes(self.name)
    
    def connect_with_host(self, host):
        """
        Connect channel with host if it is not already connected.
        Make bidirectional connection: If channel was not connected to host, 
        call also function that connects host with channel.
        """
        if host in self._connected_hosts:
            return
        self._connected_hosts.append(host)
        host.connect_with_channel(self)
        
    def connected_with_host(self, host):
        """
        Returns True if host is connected with this channel.
        """
        return host in self._connected_hosts
        
    def connect_with_process(self, process):
        """
        Connect channel with process if it is not already connected.
        Make bidirectional connection: If channel was not connected to process, 
        call also function that connects process with channel.
        """
        if process in self._connected_processes:
            return
        self._connected_processes.append(process)
        process.connect_with_channel(self)
        
    def connected_with_process(self, process):
        """
        Returns True if process is connected with this channel.
        """
        return process in self._connected_processes
    
    def is_synchronous(self):
        """
        Returns True if channel is synchronous.
        """
        return self._buffer_size == 0
    
    def has_unlimited_buffer(self):
        """
        Returns True if channel has unlimited buffer.
        """
        return self._buffer_size < 0
        
    def add_name_index(self, index):
        """
        Add index to the name of channel. 
        Before: name = ch, index = 1. After: name = ch.1
        """
        self.name += ".%d" % index
    
    def get_queue_of_receiving_hosts(self, number):
        """
        Returns list of hosts waiting for message on this channel.
        List is limited by number parameter.
        If there is less than number hosts waiting, Nones are returned 
        if channel can store messages in buffer.
        """
        hosts = []
        
        if self.is_synchronous():
            accepted_expressions_count = min([self._needed_expressions_count(), number])
            expected_messages_count = 0
            
            i = 0
            while expected_messages_count <= accepted_expressions_count and i < len(self._waiting_requests):
                request = self._waiting_requests[i]
                
                expected_messages_count += len(request.instruction.variables_names)
                
                if expected_messages_count <= accepted_expressions_count:
                    for j in range(0, len(request.instruction.variables_names)):
                        hosts.append(request.receiver)
                
                i += 1
                
        else:
            needed_messages_count = self._needed_expressions_count()
            accepted_expressions_count = number
            if not self.has_unlimited_buffer():
                accepted_expressions_count = min([self._buffer_size-len(self._sent_messages)+needed_messages_count, 
                                                  number])
    
            expected_messages_count = 0
            i = 0
            while expected_messages_count <= accepted_expressions_count:
                
                if i >= len(self._waiting_requests):
                    for j in range(expected_messages_count, accepted_expressions_count):
                        hosts.append(None)
                    break
                    
                request = self._waiting_requests[i]
                expected_messages_count += len(request.instruction.variables_names)
                if expected_messages_count <= accepted_expressions_count:
                    for j in range(0, len(request.instruction.variables_names)):
                        hosts.append(request.receiver)
                
                i += 1
                
        return hosts
    
    def get_queue_of_sending_hosts(self, number):
        """
        Return list of hosts who had sent messages to channel before.
        Works only for asynchronous channels. For synchronous channels always is returned empty array.
        Returns at most number hosts.
        """
        hosts = []
        if not self.is_synchronous():
            accepted_expressions_count = min([len(self._sent_messages), number])
            for i in range(0, accepted_expressions_count):
                message = self._sent_messages[i]
                hosts.append(message.sender)
            
        return hosts
    
    def get_number_of_dropped_messages(self):
        """
        Return number of messages dropped on this channel  
        """
        return self._dropped_messages_cnt
    
    def waiting_for_message(self, request):
        """
        Returns True if channel is waiting for message.
        """
        for i in range(0, len(self._waiting_requests)):
            waiting_request = self._waiting_requests[i]
            if waiting_request.instruction == request.instruction and waiting_request.receiver == request.receiver:
                return True
        return False
    
    def wait_for_message(self, request):
        """
        Add host to the queue of hosts waiting for messages.
        Host can wait for many expressions.
        """
        if not self.connected_with_host(request.receiver):
            raise RuntimeException("Channel '%s' is not connected with host '%s'." % (self.name, request.receiver.name))
        
        if not self.waiting_for_message(request):
            self._waiting_requests.append(request)
            
        # Check if request can be bound with expressions
        self._bind_sent_expressions_with_receivers()
    
    def _needed_expressions_count(self):
        """
        Count expressions that are needed to fulfil all waiting requests
        """
        cnt = 0
        for request in self._waiting_requests:
            cnt += len(request.instruction.variables_names)
        return cnt
    
    def send_message(self, sender_host, expressions):
        """
        Accept message with expressions.
        """
        if not self.connected_with_host(sender_host):
            raise RuntimeException("Channel '%s' is not connected with host '%s'." % (self.name, sender_host.name))
        
        if self.is_synchronous():
            
            accepted_expressions_count = min([self._needed_expressions_count(), len(expressions)])
            expected_messages_count = 0
            
            i = 0
            while expected_messages_count < accepted_expressions_count and len(self._waiting_requests) >= i:
                request = self._waiting_requests[i]
                i += 1
                
                expected_messages_count += len(request.instruction.variables_names)
                
            # Remove last addition if added to much
            if expected_messages_count > accepted_expressions_count:
                i -= 1
                request = self._waiting_requests[i]
                expected_messages_count -= len(request.instruction.variables_names)
            
            if expected_messages_count < len(expressions):
                self._dropped_messages_cnt += len(expressions) - expected_messages_count
                
            for i in range(0, expected_messages_count):
                self._sent_messages.append(ChannelMessage(sender_host, expressions[i]))
            
        else: # asynchronous
            
            # Check how many messages can be accepted.
            # If channel is not unlimited, channel can accept the minimum of: 
            # expressions sent, and buffer_size minus already waiting expressions + as many expressions as are expected at the moment
            
            accepted_expressions_cnt = len(expressions)
            if not self.has_unlimited_buffer():
                accepted_expressions_cnt = min([self._buffer_size - len(self._sent_messages) + self._needed_expressions_count(),
                                                len(expressions)])
                
            if accepted_expressions_cnt < len(expressions):
                self._dropped_messages_cnt += len(expressions) - accepted_expressions_cnt
                
                
            for i in range(0, accepted_expressions_cnt):
                self._sent_messages.append(ChannelMessage(sender_host, expressions[i]))
                
        # Check if request can be bound with expressions
        self._bind_sent_expressions_with_receivers()
            
    
class Manager():
    """ Channels manager class """
    
    def __init__(self, channels):
        self.channels = channels
        
    def find_channel_for_current_instruction(self, context):
        """
        Finds channel connected with process/host from current instruction
        and with the channel name from instruction. 
        """
        instruction = context.get_current_instruction()
        if not isinstance(instruction, CommunicationInstruction):
            return None
        return self.find_channel_for_host_instruction(context, context.get_current_host(), 
                                                      instruction)
        
    def find_channel_for_host_instruction(self, context, host, instruction):
        """
        Finds channel connected with process/host from current instruction 
        of passed host and with the channel name from instruction. 
        """
        process = host.get_current_instructions_context().get_process_of_current_list()
        if process:
            channel = context.channels_manager.find_channel_for_process(instruction.channel_name, process)
        else:
            channel = host.find_channel(instruction.channel_name)
        return channel
    
    def find_channel_for_process(self, channel_name, process):
        """
        Finds channel with name channel_name for process.
        """
        for ch in self.channels:
            if ch.original_name() == channel_name and ch.connected_with_process(process):
                return ch
        return None
    
    def build_message_request(self, receiver, communication_instruction):
        """ """
        return ChannelMessageRequest(receiver, communication_instruction)


"""
This module converts code written for Numeric to run with numpy

Makes the following changes:
 * Changes import statements (warns of use of from Numeric import *)
 * Changes import statements (using numerix) ...
 * Makes search and replace changes to:
   - .typecode()
   - .iscontiguous()
   - .byteswapped()
   - .itemsize()
   - .toscalar()
 * Converts .flat to .ravel() except for .flat = xxx or .flat[xxx]
 * Replace xxx.spacesaver() with True
 * Convert xx.savespace(?) to pass + ## xx.savespace(?)

 * Converts uses of 'b' to 'B' in the typecode-position of
   functions:
   eye, tri (in position 4)
   ones, zeros, identity, empty, array, asarray, arange,
   fromstring, indices, array_constructor (in position 2)

   and methods:
   astype --- only argument
    -- converts uses of '1', 's', 'w', and 'u' to
    -- 'b', 'h', 'H', and 'I'

 * Converts uses of type(...) is <type>
   isinstance(..., <type>)

"""
from __future__ import division, absolute_import, print_function

__all__ = ['convertfile', 'convertall', 'converttree', 'convertsrc']

import sys
import os
import re
import glob


_func4 = ['eye', 'tri']
_meth1 = ['astype']
_func2 = ['ones', 'zeros', 'identity', 'fromstring', 'indices',
         'empty', 'array', 'asarray', 'arange', 'array_constructor']

_chars = {'1':'b','s':'h','w':'H','u':'I'}

func_re = {}
meth_re = {}

for name in _func2:
    _astr = r"""(%s\s*[(][^,]*?[,][^'"]*?['"])b(['"][^)]*?[)])"""%name
    func_re[name] = re.compile(_astr, re.DOTALL)

for name in _func4:
    _astr = r"""(%s\s*[(][^,]*?[,][^,]*?[,][^,]*?[,][^'"]*?['"])b(['"][^)]*?[)])"""%name
    func_re[name] = re.compile(_astr, re.DOTALL)

for name in _meth1:
    _astr = r"""(.%s\s*[(][^'"]*?['"])b(['"][^)]*?[)])"""%name
    func_re[name] = re.compile(_astr, re.DOTALL)

for char in _chars.keys():
    _astr = r"""(.astype\s*[(][^'"]*?['"])%s(['"][^)]*?[)])"""%char
    meth_re[char] = re.compile(_astr, re.DOTALL)

def fixtypechars(fstr):
    for name in _func2 + _func4 + _meth1:
        fstr = func_re[name].sub('\\1B\\2', fstr)
    for char in _chars.keys():
        fstr = meth_re[char].sub('\\1%s\\2'%_chars[char], fstr)
    return fstr

flatindex_re = re.compile('([.]flat(\s*?[[=]))')

def changeimports(fstr, name, newname):
    importstr = 'import %s' % name
    importasstr = 'import %s as ' % name
    fromstr = 'from %s import ' % name
    fromall=0

    fstr = re.sub(r'(import\s+[^,\n\r]+,\s*)(%s)' % name,
                  "\\1%s as %s" % (newname, name), fstr)
    fstr = fstr.replace(importasstr, 'import %s as ' % newname)
    fstr = fstr.replace(importstr, 'import %s as %s' % (newname, name))

    ind = 0
    Nlen = len(fromstr)
    Nlen2 = len("from %s import " % newname)
    while True:
        found = fstr.find(fromstr, ind)
        if (found < 0):
            break
        ind = found + Nlen
        if fstr[ind] == '*':
            continue
        fstr = "%sfrom %s import %s" % (fstr[:found], newname, fstr[ind:])
        ind += Nlen2 - Nlen
    return fstr, fromall

istest_re = {}
_types = ['float', 'int', 'complex', 'ArrayType', 'FloatType',
          'IntType', 'ComplexType']
for name in _types:
    _astr = r'type\s*[(]([^)]*)[)]\s+(?:is|==)\s+(.*?%s)'%name
    istest_re[name] = re.compile(_astr)
def fixistesting(astr):
    for name in _types:
        astr = istest_re[name].sub('isinstance(\\1, \\2)', astr)
    return astr

def replaceattr(astr):
    astr = astr.replace(".typecode()", ".dtype.char")
    astr = astr.replace(".iscontiguous()", ".flags.contiguous")
    astr = astr.replace(".byteswapped()", ".byteswap()")
    astr = astr.replace(".toscalar()", ".item()")
    astr = astr.replace(".itemsize()", ".itemsize")
    # preserve uses of flat that should be o.k.
    tmpstr = flatindex_re.sub(r"@@@@\2", astr)
    # replace other uses of flat
    tmpstr = tmpstr.replace(".flat", ".ravel()")
    # put back .flat where it was valid
    astr = tmpstr.replace("@@@@", ".flat")
    return astr

svspc2 = re.compile(r'([^,(\s]+[.]spacesaver[(][)])')
svspc3 = re.compile(r'(\S+[.]savespace[(].*[)])')
#shpe = re.compile(r'(\S+\s*)[.]shape\s*=[^=]\s*(.+)')
def replaceother(astr):
    astr = svspc2.sub('True', astr)
    astr = svspc3.sub(r'pass  ## \1', astr)
    #astr = shpe.sub('\\1=\\1.reshape(\\2)', astr)
    return astr

import datetime
def fromstr(filestr):
    savestr = filestr[:]
    filestr = fixtypechars(filestr)
    filestr = fixistesting(filestr)
    filestr, fromall1 = changeimports(filestr, 'Numeric', 'numpy.oldnumeric')
    filestr, fromall1 = changeimports(filestr, 'multiarray', 'numpy.oldnumeric')
    filestr, fromall1 = changeimports(filestr, 'umath', 'numpy.oldnumeric')
    filestr, fromall1 = changeimports(filestr, 'Precision', 'numpy.oldnumeric.precision')
    filestr, fromall1 = changeimports(filestr, 'UserArray', 'numpy.oldnumeric.user_array')
    filestr, fromall1 = changeimports(filestr, 'ArrayPrinter', 'numpy.oldnumeric.array_printer')
    filestr, fromall2 = changeimports(filestr, 'numerix', 'numpy.oldnumeric')
    filestr, fromall3 = changeimports(filestr, 'scipy_base', 'numpy.oldnumeric')
    filestr, fromall3 = changeimports(filestr, 'Matrix', 'numpy.oldnumeric.matrix')
    filestr, fromall3 = changeimports(filestr, 'MLab', 'numpy.oldnumeric.mlab')
    filestr, fromall3 = changeimports(filestr, 'LinearAlgebra', 'numpy.oldnumeric.linear_algebra')
    filestr, fromall3 = changeimports(filestr, 'RNG', 'numpy.oldnumeric.rng')
    filestr, fromall3 = changeimports(filestr, 'RNG.Statistics', 'numpy.oldnumeric.rng_stats')
    filestr, fromall3 = changeimports(filestr, 'RandomArray', 'numpy.oldnumeric.random_array')
    filestr, fromall3 = changeimports(filestr, 'FFT', 'numpy.oldnumeric.fft')
    filestr, fromall3 = changeimports(filestr, 'MA', 'numpy.oldnumeric.ma')
    fromall = fromall1 or fromall2 or fromall3
    filestr = replaceattr(filestr)
    filestr = replaceother(filestr)
    if savestr != filestr:
        today = datetime.date.today().strftime('%b %d, %Y')
        name = os.path.split(sys.argv[0])[-1]
        filestr = '## Automatically adapted for '\
                  'numpy.oldnumeric %s by %s\n\n%s' % (today, name, filestr)
        return filestr, 1
    return filestr, 0

def makenewfile(name, filestr):
    fid = file(name, 'w')
    fid.write(filestr)
    fid.close()

def convertfile(filename, orig=1):
    """Convert the filename given from using Numeric to using NumPy

    Copies the file to filename.orig and then over-writes the file
    with the updated code
    """
    fid = open(filename)
    filestr = fid.read()
    fid.close()
    filestr, changed = fromstr(filestr)
    if changed:
        if orig:
            base, ext = os.path.splitext(filename)
            os.rename(filename, base+".orig")
        else:
            os.remove(filename)
        makenewfile(filename, filestr)

def fromargs(args):
    filename = args[1]
    converttree(filename)

def convertall(direc=os.path.curdir, orig=1):
    """Convert all .py files to use numpy.oldnumeric (from Numeric) in the directory given

    For each changed file, a backup of <usesnumeric>.py is made as
    <usesnumeric>.py.orig.  A new file named <usesnumeric>.py
    is then written with the updated code.
    """
    files = glob.glob(os.path.join(direc, '*.py'))
    for afile in files:
        if afile[-8:] == 'setup.py': continue # skip these
        convertfile(afile, orig)

header_re = re.compile(r'(Numeric/arrayobject.h)')

def convertsrc(direc=os.path.curdir, ext=None, orig=1):
    """Replace Numeric/arrayobject.h with numpy/oldnumeric.h in all files in the
    directory with extension give by list ext (if ext is None, then all files are
    replaced)."""
    if ext is None:
        files = glob.glob(os.path.join(direc, '*'))
    else:
        files = []
        for aext in ext:
            files.extend(glob.glob(os.path.join(direc, "*.%s" % aext)))
    for afile in files:
        fid = open(afile)
        fstr = fid.read()
        fid.close()
        fstr, n = header_re.subn(r'numpy/oldnumeric.h', fstr)
        if n > 0:
            if orig:
                base, ext = os.path.splitext(afile)
                os.rename(afile, base+".orig")
            else:
                os.remove(afile)
            makenewfile(afile, fstr)

def _func(arg, dirname, fnames):
    convertall(dirname, orig=0)
    convertsrc(dirname, ext=['h', 'c'], orig=0)

def converttree(direc=os.path.curdir):
    """Convert all .py files and source code files in the tree given
    """
    os.path.walk(direc, _func, None)


if __name__ == '__main__':
    fromargs(sys.argv)
